;;;  -*- mode: LISP; Package: CL-USER; Syntax: COMMON-LISP;  Base: 10 -*-
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Author      : Mike Byrne
;;; Copyright   : (c)2000-3 Rice U./Mike Byrne, All Rights Reserved
;;; Availability: public domain
;;; Address     : Rice University
;;;             : Psychology Department
;;;             : Houston,TX 77005-1892
;;;             : byrne@acm.org
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Filename    : emma.lisp
;;; Version     : r11
;;;
;;; Description : Implementation of Dario Salvucci's EMMA system for eye
;;;             : movements based on subclassing the Vision Module and making
;;;             : eye movements based on movement styles.
;;;
;;; Bugs        : "Chunks created early" problem.
;;;
;;; Todo        : [] Solve "chunks created early" problem [requires some
;;;             :     rewriting of the Vision Module].
;;;             : [] Tracking.
;;;
;;; ----- History -----
;;; 00.07.12 Mike Byrne
;;;             :  Incept date.
;;; 01.05.09 mdb [r2]
;;;             : Fixed vector bugs.  Thanks, Dario.
;;; 01.05.?? Dario Salvucci [r3]
;;;             : Noise added to both X and Y of movement.
;;; 01.05.28 mdb [r4]
;;;             : Some minor tinkering with Dario's last changes.
;;; 01.07.02 mdb [r5]
;;;             : Made eye movement relative to next landing spot.
;;; 01.09.21 mdb [r6]
;;;             : Fixed bug in COMPLETE-EYE-MOVE.
;;; 01.10.01 mdb [r7]
;;;             : Added PROC-S method that's actually correct.
;;; 01.11.10 mdb
;;;             : Changed trace log to use CONS rather than LIST to save
;;;             : a smidge of memory, and made tracing the eye path switchable.
;;; 2002.05.04 [r8]
;;;             : Removed unnecessary PROC-STUB stuff.
;;; 2002.05.07 
;;;             : Fixed problem of FIND-LOCATION being clobbered by new preps.
;;;             : Renamed FOCUS-ON to ENCODING-COMPLETE.
;;; 2003.05.07 mdb [r9]
;;;             : Fixed MCL-specific checking. 
;;; 2003.06.23 mdb [r10]
;;;             : Made it easier and cleaner to hook into eye spot updating 
;;;             : for non-MCL Lisps by making a more generic function call 
;;;             : to which specific devices can be tailored, paralleling how
;;;             : it's done for attended location. 
;;;             : Based the EYE-SPOT class on the new RPM-OVERLAY class from
;;;             : the 2.1.3 version of "mcl-device.lisp".
;;; 2003.07.17 mdb [r11]
;;;             : Added a CLEAR :after method to kill any in-progress shift
;;;             : and stripped some dead code.
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(defvar *eye-spot* nil)


(defclass emma-vis-mod (vision-module)
  ((freq-ht :accessor freq-ht :initarg :freq-ht
            :initform (make-hash-table :test #'equal))
   (enc-factor :accessor enc-factor :initarg :enc-factor :initform 0.010)
   (enc-exponent :accessor enc-exponent :initarg :enc-exponent :initform 1.0)
   (eye-loc :accessor eye-loc :initform #(0 0))
   (eye-trace :accessor eye-trace :initform nil)
   (base-exe :accessor base-exe :initarg :base-exe :initform 0.020)
   (sacc-rate :accessor sacc-rate :initarg :sacc-rate :initform 0.002)
   (shift-start :accessor shift-start :initarg :shift-start :initform 0)
   (shift-duration :accessor shift-duration :initarg :shift-duration
                   :initform nil)
   (shift-target :accessor shift-target :initarg :shift-target :initform nil)
   (trace-eye-p :accessor trace-eye-p :initarg :trace-eye-p :initform nil)
   ))


(defmethod reset-module :after ((eye-mod emma-vis-mod))
  (setf (eye-trace eye-mod) nil)
  (set-eye-loc eye-mod #(0 0))
  (setf (shift-start eye-mod) 0.0)
  (setf (shift-target eye-mod) nil)
  (clrhash (freq-ht eye-mod))
  )

(defmethod set-eye-loc ((eye-mod emma-vis-mod) (newloc vector))
  (setf (eye-loc eye-mod) newloc)
  (device-update-eye-loc (device (device-interface *mp*)) newloc)
  (when (trace-eye-p eye-mod)
    (push (cons (pm-time) newloc) (eye-trace eye-mod))))


;;;; ---------------------------------------------------------------------- ;;;;
;;;; Saccade movement style

(defStyle saccade () r theta trgt)


;;; COMPUTE-EXEC-TIME      [Method]
;;; Date        : 01.05.09
;;; Description : Execution time is the sum of the base execution time and
;;;             : the distance times the rate.

(defmethod compute-exec-time ((eye-mod emma-vis-mod) (mvmt saccade))
  (rand-time
   (+ (init-time eye-mod) (* (r mvmt) (sacc-rate eye-mod)) 
      (base-exe eye-mod))))


;;; COMPUTE-FINISH-TIME      [Method]
;;; Date        : 01.05.09
;;; Description : The finish time is the same as the execution time for
;;;             : saccades.

(defmethod compute-finish-time ((eye-mod emma-vis-mod) (mvmt saccade))
  (exec-time mvmt))


;;; FEAT-DIFFERENCES      [Method]
;;; Date        : 01.05.09
;;; Description : How many features to prepare?  One for direction and one for
;;;             : distance, if they're different.

(defmethod feat-differences ((s1 saccade) (s2 saccade))
  (let ((nfeats 0))
    (unless (distance= (r s1) (r s2)) (incf nfeats))
    (unless (direction= (theta s1) (theta s2)) (incf nfeats))
    nfeats))


;;; TOTAL-TIME      [Method]
;;; Date        : 01.05.09
;;; Description : The total time for a saccade is the simple sum of the
;;;             : prep time and the execution time.

(defmethod total-time ((sacc saccade))
  (+ (fprep-time sacc) (exec-time sacc)))


;;; PERFORM-MOVEMENT      [Method]
;;; Date        : 01.05.09
;;; Description : Performing the movement simply involves figuring out when
;;;             : to call COMPLETE-EYE-MOVE and updating the eye location.

(defmethod perform-movement ((eye-mod emma-vis-mod) (sacc saccade))
  (change-state eye-mod :exec 'BUSY)
  (let ((new-loc (polar-move-xy (eye-loc eye-mod)
                                  (vector (pm-angle-to-pixels (r sacc))
                                          (theta sacc)))))
    (queue-command :command 'COMPLETE-EYE-MOVE :time (exec-time sacc)
                   :randomize nil :where :VISION
                   :params `(,(trgt sacc) ,new-loc))))




;;;; ---------------------------------------------------------------------- ;;;;
;;;; Module-level stuff





;;; MOVE-ATTENTION      [Method]
;;; Date        : 01.05.09
;;; Description : Only slightly more complicated than without EMMA.  First,
;;;             : clear all the bookkeeping stuff [e.g. where attention is].
;;;             : Unless we're in the middle of execution, clear the module's
;;;             : input queue.  This will effectively destroy any previously-
;;;             : prepared movement, as it will eliminate any
;;;             : PREPARATION-COMPLETE entries in the queue.  Finally, we
;;;             : need to set up an initiate the actual eye move.
;;;             : Note that the computation of the recogntion time for the
;;;             : target assumes that the object that is there when the move
;;;             : starts and ends will be the same.  Thus, this may not work
;;;             : quite right for very dynamic displays.

(defmethod move-attention ((eye-mod emma-vis-mod) &key location scale)
  (declare (symbol scale))
  ;; old stuff
  (when (tracked-obj eye-mod) (remove-tracking eye-mod))
  (setf (moving-attention eye-mod) (gentemp "TRGT"))
  (clear-attended eye-mod)
  (setf (last-scale eye-mod) scale)
  (setf location (psdme-to-dmo location))
  (setf (current-marker eye-mod) location)
  (set-clof eye-mod (dmo-to-xy location))
  ;; new EMMA stuff. New prep clears old prep.
  (unless (eq 'BUSY (exec-s eye-mod))
    (setf (input-q eye-mod) 
          (mklist (first (member 'FIND-LOCATION (input-q eye-mod) 
                                 :key #'(lambda (f)
                                          (first (params f)))))))    
    )
  ;; need the object there to compute the recognition time
  (let ((return-obj (get-obj-at-location eye-mod location scale)))
    (when return-obj
      (let* ((start-loc (next-destination eye-mod))
             (end-loc (dmo-to-xy location))
             (r-theta (compute-saccade-r-theta start-loc end-loc)))
        (change-state eye-mod :proc 'BUSY)
        (setf (shift-duration eye-mod) (recog-time eye-mod return-obj r-theta))
        (setf (shift-target eye-mod) return-obj)
        (initiate-eye-move eye-mod (shift-duration eye-mod) r-theta)))))



(defgeneric next-destination (eye-mod)
  (:documentation  "Returns the next landing location for EMMA."))

(defmethod next-destination ((eye-mod emma-vis-mod))
  (if (eq 'FREE (exec-s eye-mod))
    (eye-loc eye-mod)
    (third (params (find 'COMPLETE-EYE-MOVE (input-q eye-mod)
                         :key #'(lambda (e) (first (params e))))))
    ))


;;; INITIATE-EYE-MOVE      [Method]
;;; Date        : 01.05.09
;;; Description : First, build the actual saccade movement itself.  This is a
;;;             : CLOS object whose class is generated by the DEFSTYLE call in
;;;             : this file.  Do all the state changes and queue up the
;;;             : preparation complete notice, which will then generate an
;;;             : eye movement.  Now, we only want to queue the FOCUS-ON when
;;;             : the object will be recognized before the eye movement is
;;;             : complete, so do that check and queue that event if necessary.

(defgeneric initiate-eye-move (eye-mod recog-time r-theta)
  (:documentation "Given a recognition time and a movement spec, begin an eye movement."))

(defmethod initiate-eye-move ((eye-mod emma-vis-mod) (recog-time number)
                                 (r-theta vector))
  (let ((sacc-mvmt (build-sacc-mvmt eye-mod r-theta)))
    ;; always start the preparation of the eye move
    (setf (shift-start eye-mod) (pm-time))
    (change-state eye-mod :prep 'BUSY)
    (setf (last-prep eye-mod) sacc-mvmt)
    (queue-command :command 'PREPARATION-COMPLETE :where :VISION
                   :time (fprep-time sacc-mvmt) :randomize nil)
    ;; only sometimes queue the FOCUS-ON
    (when (<= recog-time (total-time sacc-mvmt))
      (queue-command
       :command 'ENCODING-COMPLETE :where :VISION
       :time recog-time :randomize nil
       :params `(,(current-marker eye-mod) ,(last-scale eye-mod))))))


(defmethod initiate-eye-move ((eye-mod emma-vis-mod) (recog-time number)
                                 (r-theta list))
  (initiate-eye-move eye-mod recog-time (coerce r-theta 'vector)))



;;; COMPLETE-EYE-MOVE      [Method]
;;; Date        : 01.05.09
;;; Description : When an eye movement completes:
;;;             : [1] Update the blue POR spot on the display, the eye trace,
;;;             :     and the module's state.
;;;             : [2] Compute the new recognition time for the target based
;;;             :     on the new distance.
;;;             : [3] Initiate an eye move.  Note that this movement should
;;;             :     die if recog time based on the eye's new location is
;;;             :     small.


(defgeneric complete-eye-move (eye-mod marker xyloc)
  (:documentation "Called when an eye movement completes to handle all the updating."))

(defmethod complete-eye-move ((eye-mod emma-vis-mod) (trgt symbol) 
                                 (xyloc vector))
  ;; handle the module-level stuff
  (set-eye-loc eye-mod xyloc)
  ;; when we're still moving attention to the same target, we might need
  ;; to generate another shift.
  (when (and (moving-attention eye-mod)
             (eq trgt (moving-attention eye-mod)))
    (let* ((start-loc (eye-loc eye-mod))
           (end-loc (dmo-to-xy (current-marker eye-mod)))
           (r-theta (compute-saccade-r-theta start-loc end-loc))
           (new-duration
            (* (recog-time eye-mod (shift-target eye-mod) r-theta)
               (- 1 (/ (- (pm-time) (shift-start eye-mod))
                       (shift-duration eye-mod))))))
      (setf (shift-start eye-mod) (pm-time))
      (setf (shift-duration eye-mod) new-duration)
      (initiate-eye-move eye-mod new-duration r-theta)))
  (finish-movement eye-mod)
  )

(defmethod complete-eye-move ((eye-mod emma-vis-mod) (trgt symbol) 
                                 (xyloc list))
  (complete-eye-move eye-mod trgt (coerce xyloc 'vector)))



(defgeneric compute-saccade-r-theta (start-loc end-loc)
  (:documentation "Adds gaussian noise to <end-loc> and returns r-theta vector for saccade destination."))

(defmethod compute-saccade-r-theta ((start-loc vector) (end-loc vector))
  (let* ((stddev (* 0.1 (dist start-loc end-loc)))
         (end-loc-noisy 
          (vector (add-gaussian-noise (aref end-loc 0) stddev)
                  (add-gaussian-noise (aref end-loc 1) stddev))))
    (xy-to-polar start-loc end-loc-noisy)))




(defgeneric recog-time (eye-mod obj r-theta)
  (:documentation "Based on Dario's equation, compute the recognition 
time for an object given the displacement from the current POR."))

(defmethod recog-time ((eye-mod emma-vis-mod) (obj dmo) (r-theta vector))
  (let ((freq (object-frequency eye-mod obj))
        (eccentricity (pm-pixels-to-angle (vr r-theta))))
    (rand-time (* (enc-factor eye-mod)
                  (- (log freq))
                  (exp (* eccentricity (enc-exponent eye-mod)))))))

(defmethod recog-time ((eye-mod emma-vis-mod) (obj dmo) (r-theta list))
  (recog-time eye-mod obj (coerce r-theta 'vector)))



;;; OBJECT-FREQUENCY      [Method]
;;; Description : Implicit default of 0.01 if there's no hash table entry
;;;             : or feature-based frequency on the object.

(defgeneric object-frequency (eye-mod obj &optional default)
  (:documentation "Compute the frequency of ocurrence of a visual 
object, which is used to compute the recognition time."))


(defmethod object-frequency ((eye-mod emma-vis-mod) (obj dmo)
                                &optional (default 0.01))
  (aif (stringp (get-attribute obj 'value))
    (aif (gethash it (freq-ht eye-mod))
      it
      default)
    (aif (feat-named eye-mod (id obj))
      (obj-freq it)
      default)))


(defgeneric build-sacc-mvmt (eye-mod r-theta)
  (:documentation "Constructs a saccade movement based on a displacement from the current POR."))


(defmethod build-sacc-mvmt ((eye-mod emma-vis-mod) (r-theta vector))
  (let ((sacc-mvmt (make-instance 'saccade
                       :r (pm-pixels-to-angle (vr r-theta))          ; removed noise addition 01.05.28
                       :theta (vtheta r-theta))))
    (setf (fprep-time sacc-mvmt)
          (rand-time (compute-prep-time eye-mod sacc-mvmt)))
    (setf (exec-time sacc-mvmt)
          (compute-exec-time eye-mod sacc-mvmt))
    (setf (trgt sacc-mvmt) (moving-attention eye-mod))
    sacc-mvmt))

(defmethod build-sacc-mvmt ((eye-mod emma-vis-mod) (r-theta list))
  (build-sacc-mvmt eye-mod (coerce r-theta 'vector)))


(defgeneric register-obj-freq (eye-mod name freq)
  (:documentation "Store a frequency for a particular symbol."))

(defmethod register-obj-freq ((eye-mod emma-vis-mod) (name string)
                                 (freq float))
  (setf (gethash name (freq-ht eye-mod)) freq))


;;; FOCUS-ON      [Method]
;;; Date        : 01.05.10
;;; Description : This is essentially the same as FOCUS-ON for the non-EMMA
;;;             : Vision Module, but it also checks whether the target of
;;;             : the attention shift is still there.

(defmethod encoding-complete ((eye-mod emma-vis-mod) (loc-dmo dmo) scale)
  (declare (symbol scale))
  (set-attributes loc-dmo '(attended t))
  (setf (moving-attention eye-mod) NIL)
  (change-state eye-mod :proc 'FREE)
  (let ((return-obj
         (if (and (shift-target eye-mod)
                  (object-present-p eye-mod (id (shift-target eye-mod))))
           (shift-target eye-mod)
           (get-obj-at-location eye-mod loc-dmo scale))))
    (unless return-obj
      ;(clear-loc loc-dmo)
      (clear-attended eye-mod)
      (return-from encoding-complete nil))
    (set-attended eye-mod return-obj)
    (when (print-view-p eye-mod)
      (format t "~% <> MOVE-ATTENTION generated DMO ~S~%" (id return-obj)))
    return-obj))


;;; UPDATE-ATTENDED-LOC      [Method]
;;; Date        : 01.05.10
;;; Description : Also much like the non-EMMA method. If the screen at the
;;;             : attended location changes, there's some updating to do
;;;             : and an attention shift to the new object, if any.

(defmethod update-attended-loc ((eye-mod emma-vis-mod))
  (when (and (current-marker eye-mod)
             (null (tracked-obj eye-mod))
             (not (moving-attention eye-mod)))
    (let ((new-obj (get-obj-at-location eye-mod (current-marker eye-mod)
                                          (last-scale eye-mod))))
      (when (and new-obj
                 (currently-attended eye-mod)
                 (not (object-present-p eye-mod
                                        (id (currently-attended eye-mod)))))
        (move-attention eye-mod
                        :location (dmo-to-psdme (current-marker eye-mod))
                        :scale (last-scale eye-mod))))))


#|
        (queue-command
         :time (move-attn-latency vis-mod)
         :where :VISION
         :command 'focus-on
         :randomize t
         :params `(,(current-loc vis-mod) ,(last-scale vis-mod)))
        (change-state vis-mod :exec 'busy))))
|#


(defmethod clear :after ((eye-mod emma-vis-mod))
  (setf (moving-attention eye-mod) nil))



;;;; ---------------------------------------------------------------------- ;;;;
;;;; Utilities


(defun add-gaussian-noise (x stddev)
  "Adds pseudo-Gaussian noise to mean x with a given stddev."
  (let* ((v (* stddev stddev))
         (s (/ (sqrt (* 3.0 v)) pi)))
    (+ x (noise s))))


(defgeneric device-update-eye-loc (device xyloc)
  (:documentation "Called by EMMA when the eye position changes so it can be rendered appropriately."))


;;; DEVICE-UPDATE-EYE-LOC      [Method]
;;; Description : The base method for updating does nothing.  Redefine for
;;;             : your device--there is a method for MCL below.

(defmethod device-update-eye-loc (device xyloc)
  (declare (ignore device xyloc))
  nil)




;;;; ---------------------------------------------------------------------- ;;;;
;;;; tracking stuff  !!MCL-specific!!

#+(and :mcl (not :openmcl))
(progn
  
  (defclass eye-spot (rpm-overlay)
    ((color :accessor color :initarg :color :initform *blue-color*))
    (:default-initargs 
      :view-size #@(9 9)
      :offset #@(-5 -5)))
  
  (defmethod view-draw-contents ((self eye-spot))
    (with-focused-view self
      (with-fore-color (color self)
        (fill-oval self *light-gray-pattern* #@(0 0) (view-size self)))))
  
  
  (defmethod device-update-eye-loc ((device window) (xyloc vector))
    (update-me *eye-spot* device xyloc))
  
  (setf *eye-spot* (make-instance 'eye-spot))
  )

;;;; ---------------------------------------------------------------------- ;;;;
;;;; toplevel stuff

(defun pm-register-string-frequency (string frequency)
  "Register the frequency of a string."
  (register-obj-freq (vis-m *mp*) string frequency))



#|
(defun pm-eye-movement-trace ()
  "Returns a trace of the model's eye movements as a list of (time (x y))."
  (reverse (eye-trace (vis-m *mp*))))
|#

;;; temporary wrapper

(defun pm-add-visual-object (name freq)
  (pm-register-string-frequency name freq))


(defun pm-set-eye-location (loc)
  (set-eye-loc (vis-m *mp*) (coerce loc 'vector)))

(defun pm-current-eye-location ()
  (eye-loc (vis-m *mp*)))



;;;; ---------------------------------------------------------------------- ;;;;
;;;; Parameters


#|  Now included in "patch-support.lisp"
(defun add-parameter (param)
  "Add a PM parameter."
  (when (assoc (name param) (param-lis *mp*))
    (delete-parameter (name param)))
  (push (list (name param) param) (param-lis *mp*))
  (setf (param-lis *mp*)
        (sort (param-lis *mp*) #'key-lessp :key #'first)))


(defun delete-parameter (name)
  "Deletes a PM parameter."
  (setf (param-lis *mp*)
        (delete name (param-lis *mp*) :key 'first)))


(defun key-lessp (k1 k2)
  (string-lessp (mkstr k1) (mkstr k2)))
|#


(mapc #'add-parameter
      (list
       (make-instance 'pm-param :name :VISUAL-ENCODING-FACTOR
                      :test #'posnum
                      :place '(enc-factor (vis-m *mp*))
                      :default 0.010)
       (make-instance 'pm-param :name :VISUAL-ENCODING-EXPONENT
                      :test #'posnum
                      :place '(enc-exponent (vis-m *mp*))
                      :default 1.0)
       (make-instance 'pm-param :name :EYE-SACCADE-RATE
                      :test #'posnum
                      :place '(sacc-rate (vis-m *mp*))
                      :default 0.002)
       (make-instance 'pm-param :name :SACCADE-BASE-TIME
                      :test #'posnum
                      :place '(base-exe (vis-m *mp*))
                      :default 0.020)
       ))





;;;; ---------------------------------------------------------------------- ;;;;
;;;; Init stuff

(if (fboundp 'pm-install-module)
  (pm-install-module *mp* (make-instance 'emma-vis-mod) nil)
  (setf (vis-m *mp*) (make-instance 'emma-vis-mod)))


(setf vm (vis-m *mp*))

